#!/bin/perl

# 3dstoobj -- convert a .3ds file to wavefront .obj format
#   
# usage:  3dstoobj [-ignorematrix] [-offset=nnnn] file.3ds
#    -ignorematrix -- some 3ds files have "invalid" matrix chunks
#    -offset       -- debugging mostly...
#    -v            -- verbose
#
# 09/15/96 chas williams <chas@ohm.nrl.navy.mil>
#          first release

$opt'in = "";
$opt'out = "";
#$opt'ignorematrix = 0;
$opt'offset = 0;
$opt'verbose = 0;

while ( $_ = shift )
{
   case:
   {

      /^-offset/ && (($opt'offset) = /offset=(\d+)/, last case);
      /^-v/ && (++$opt'verbose, last case);
      #/^-ignorematrix/ && ( $opt'ignorematrix = 1, last case);
      /-d/ && do 
      {
         printf STDERR "3dstoobj: bad option: \"%s\"\n", $_;
         last case;
      };

      $opt'in = $_;
   }
}

open(DDD,"$opt'in") ||
   die "3dstoobj: failed to open $opt'in: $!\n";

if ( $opt'offset != 0 )
   { seek(DDD, $opt'offset, 1) };	# skip some "header" bytes

$opt'out = $opt'in;
$opt'out =~ s/\.(\w+)$//;	# remove extension
$opt'out = "$opt'out.obj";	# append .obj

open(OBJ,">$opt'out") ||
   die "3dstoobj: failed to open $opt'out: $!\n";

$obj'_offset = 0;

$ddd'_depth = 0;

read(DDD, $chunk, 2);
$chunk = unpack("H4", pack("S", unpack("v", $chunk)));
if ( $chunk eq "4d4d" )
{
   read(DDD, $size, 4);
   $size = unpack("V", $size);
   &handle_chunks($size-6);
}
else
{
   printf STDERR "3dstoobj: input doesn't appear to be a 3DS file\n"; 
   printf STDERR "3dstoobj: try using -offset=\n"; 
}

close(DDD);
close(OBJ);

exit;

sub
handle_chunks
{
   local($chunksize) = @_;

   ++$ddd'_depth;

   while( $chunksize > 0 )
   {
      # argh! little endian! its making my life hell...
      read(DDD, $chunk, 2) ||
         die "3dstoobj: read of chunk id failed: $chunksize $!\n";
      read(DDD, $size, 4) ||
         die "3dstoobj: read of chunk size failed: $!\n";

      $chunk = unpack("H4", pack("S", unpack("v", $chunk)));
      $size = unpack("V", $size);

      $chunksize -= $size;

      # debugging
      #printf "chunk= $chunk\n";
      #printf "size=  $size\n";

      case:
      {
         $chunk eq "4d4d" && (&ddd'init, last case);
		 $chunk eq "0002" && ( &ddd'version(), last case);

   	 $chunk eq "3d3d" && (&ddd'objects($size), last case);
   	    $chunk eq "3d3e" && (&ddd'meshversion, last case);

   	    $chunk eq "afff" && ( &ddd'materialdef($size), last case);
   		 $chunk eq "a000" && ( &ddd'materialname(), last case);
   		 $chunk eq "a010" && ( &ddd'ambient($size), last case);
			 $chunk eq "0011" && ( &ddd'color24(), last case);
   		 $chunk eq "a020" && ( &ddd'diffuse($size), last case);
   		 $chunk eq "a030" && ( &ddd'specular($size), last case);
   		 $chunk eq "a040" && ( &ddd'shininess($size), last case);
			 $chunk eq "0031" && ( &ddd'floatpercent, last case);

   	 $chunk eq "4000" && (&ddd'object($size), last case);
   	    $chunk eq "4100" && (&ddd'triangles($size), last case);
   	       $chunk eq "4110" && (&ddd'points, last case);
   	       $chunk eq "4111" && (&ddd'pflags, last case);
   	       $chunk eq "4120" && (&ddd'faces($size), last case);
   	          $chunk eq "4130" && (&ddd'material, last case);
   	       $chunk eq "4160" && (&ddd'matrix, last case);
   	       $chunk eq "4165" && (&ddd'color, last case);


         # skip this "unknown" chunk
         printf STDERR "%s%s (unknown (size=%d))\n",
           "\t" x $ddd'_depth, $chunk, $size;
         $opt'verbose > 1 && do
         {
   	    while( $size-6 > 15 )
   	    {
               read(DDD, $buf, 16);
               $hex = unpack('h32', $buf);
               $buf =~ tr/\0-\37\177-\377/./;
               printf STDERR "\t%s %s\n", $hex, $buf;
               $size -= 16;
            }
            read(DDD, $buf, $size-6);
            $hex = unpack('h32', $buf);
            $buf =~ tr/\0-\37\177-\377/./;
            printf STDERR "\t%s %s\n", $hex, $buf;
            $size = 6;
         };

         seek(DDD, $size-6, 1);

      }
   }

   --$ddd'_depth;
}

sub
ddd'init # 4d4d
{
   printf STDERR "%s (3ds file)\n", $chunk;
}

sub
ddd'objects
{
   local($size) = @_;
   printf STDERR "%s (objects size=%d)\n", $chunk, $size;
   &handle_chunks($size-6);
}

sub
ddd'object # 4000 -- object (contains subchunks)
{
   local($size) = @_;

   undef @matrix;
   undef @points;
   undef @faces;
   undef @material;

   $object = &ddd'read_string();

   printf STDERR "%s%s (name=\"%s\" %d)\n",
           "\t" x $ddd'_depth, $chunk, $object, $size;

   $size -= length($object)+1;

   &handle_chunks($size-6);

   # ok, the object should be entirely in memory now
   printf OBJ "g %s\n", $object;

   $obj'_offset = $obj'npts + 1;		# offset into .obj point list

   for($i=0; $i<$npoints; ++$i)
   {
      @p = split(',', $points[$i]);
      printf OBJ "v %f %f %f\n", $p[0], $p[1], $p[2];
   }
   $obj'npts += $npoints;

   $lastmaterial = "";

   for($i=0; $i<$nfaces; ++$i)
   {
      if (defined($material[$i]) & $lastmaterial ne $material[$i])
      {
         printf OBJ "usemtl %s\n", $material[$i];
         $lastmaterial = $material[$i];
      }

      @p = split(',', $faces[$i]);
      printf OBJ "f %d %d %d\n",
         $p[2]+$obj'_offset, $p[1]+$obj'_offset, $p[0]+$obj'_offset;
   }

}

sub
ddd'triangles # 4100 -- triangles (contains subchunks)
{
   local($size) = @_;

   printf STDERR "%s%s (triangles %d)\n",
           "\t" x $ddd'_depth, $chunk, $size;

   &handle_chunks($size-6);
}

sub
ddd'points # 4110 -- list of points
{
   $npoints = &ddd'read_short();

   printf STDERR "%s%s (npoints=%d)\n",
           "\t" x $ddd'_depth, $chunk, $npoints;

   for($i=0; $i<$npoints; ++$i)
   {
      $x = &ddd'read_float();
      $y = &ddd'read_float();
      $z = &ddd'read_float();

      $points[$i] = "$x,$y,$z";
   }

}

sub
ddd'faces # 4120 -- faces (contains subchunks)
{
   local($size) = @_;

   $nfaces = &ddd'read_short();

   printf STDERR "%s%s (nfaces=%d)\n",
           "\t" x $ddd'_depth, $chunk, $nfaces;


   for($i=0; $i<$nfaces; ++$i)
   {
      $p[0] = &ddd'read_short();
      $p[1] = &ddd'read_short();
      $p[2] = &ddd'read_short();
      $flags = &ddd'read_short();	# ???

      $faces[$i] = "$p[0],$p[1],$p[2]";
   }

   $size -= 2+($nfaces*8);

   &handle_chunks($size-6);
}

sub
ddd'version
{
   read(DDD, $version, 4);

   $version = unpack("V", $version);

   printf STDERR "%s (version=%d)\n", $chunk, $version;
}

sub
ddd'meshversion
{
   read(DDD, $version, 4);
   $version = unpack("V", $version);
   printf STDERR "%s%s (meshversion=%d)\n",
      "\t" x $ddd'_depth, $chunk, $version;
}

sub
ddd'pflags # 4111 -- points flags ???
{

   printf STDERR "%s%s (pflags ???)\n",
           "\t" x $ddd'_depth, $chunk;

   #read(DDD, $n, 2);
   #$n = unpack("v", $n);

   seek(DDD, $size-6, 1); # skipit
}

sub
ddd'matrix # 4160
{

   printf STDERR "%s%s (matrix)\n",
           "\t" x $ddd'_depth, $chunk;

   for($i=0; $i<4; ++$i)
   {
      for($j=0; $j<3; ++$j)
      {
         read(DDD, $f, 4);
         $matrix[($i*4)+$j] = unpack("f", pack("L", unpack("V", $f)));
      }
   }

   # 3ds assumes the last row of a 4x4 translation matrix is 0001
   $matrix[(0*4)+3] = 0;
   $matrix[(1*4)+3] = 0;
   $matrix[(2*4)+3] = 0;
   $matrix[(3*4)+3] = 1;

   $opt'verbose > 1 && do
   {
      for($i=0; $i<4; ++$i)
      {
         printf STDERR "\t\t[";
         for($j=0; $j<4; ++$j)
            { printf STDERR " %7.3f", $matrix[($i*4)+$j] };
         printf STDERR " ]\n";
      }
   };

}

sub
ddd'materialdef # afff -- material definition (contains subchunks)
{
   local($size) = @_;

   printf STDERR "%s%s (materialdef)\n",
      "\t" x $ddd'_depth, $chunk;

   &handle_chunks($size-6);
}

sub
ddd'materialname # a000 -- material name
{
   $name = "";
   while( read(DDD, $c, 1) )
   {
      last if $c eq "\000";
      $name = "$name$c";
   }

   printf STDERR "%s%s (name=%s)\n",
      "\t" x $ddd'_depth, $chunk, $name;
}

sub
ddd'ambient # a010 -- material ambient (contains subchunks)
{
   local($size) = @_;

   printf STDERR "%s%s (ambient)\n",
      "\t" x $ddd'_depth, $chunk;

   &handle_chunks($size-6);
}

sub
ddd'diffuse # a020 -- material diffuse (contains subchunks)
{
   local($size) = @_;

   printf STDERR "%s%s (diffuse)\n",
      "\t" x $ddd'_depth, $chunk;

   &handle_chunks($size-6);
}

sub
ddd'specular # a030 -- material specular (contains subchunks)
{
   local($size) = @_;

   printf STDERR "%s%s (specular)\n",
      "\t" x $ddd'_depth, $chunk;

   &handle_chunks($size-6);
}

sub
ddd'shininess # a040 -- material shininess (contains subchunks)
{
   local($size) = @_;

   printf STDERR "%s%s (shininess)\n",
      "\t" x $ddd'_depth, $chunk;

   &handle_chunks($size-6);
}

sub
ddd'color24 # 0011 -- color specification
{
   read(DDD, $r, 1);
   $r = unpack("C", $r);
   read(DDD, $g, 1);
   $g = unpack("C", $g);
   read(DDD, $b, 1);
   $b = unpack("C", $b);

   printf STDERR "%s%s (color=[%d,%d,%d])\n",
      "\t" x $ddd'_depth, $chunk, $r, $g, $b;
}

sub
ddd'floatpercent # 0031 -- percentage (float)
{
   read(DDD, $percent, 4);
   $percent = unpack("f", pack("L", unpack("V", $percent)));

   printf STDERR "%s%s (percent %g)\n",
      "\t" x $ddd'_depth, $chunk, $percent;
}

sub
ddd'material # 4130 -- material for faces
{
   $material = &ddd'read_string();
   $n = &ddd'read_short();

   printf STDERR "%s%s (\"%s\" (%d faces))\n",
      "\t" x $ddd'_depth, $chunk, $material, $n;

   $material =~ tr/ /_/;	# convert to acceptable .obj material names

   for($i=0; $i<$n; ++$i)
   {
      $face = &ddd'read_short();
      
      $material[$face] = $material;
   }
}

sub
ddd'color # 4165
{
   read(DDD, $n, 1);
   $n = unpack("b", $n);

   printf STDERR "\t%s (color=%d--ignored)\n", $chunk, $n;
}

sub
matmul
{
   local(@a[0..3], @b[0..15]) = @_;
   local(@c, $i, $j);

   for ($i=0; $i<4; ++$i)
   {
      for ($j=0; $j<4; ++$j)
      {
         $c[$i] += $a[$i]*$b[(4*$j)+$i];
      }
   }

   return @c;
}

sub
ddd'read_short
{
   local($i);
   read(DDD, $i, 2);
   return unpack("v", $i);
}

sub
ddd'read_float
{
   local($f);
   read(DDD, $f, 4);
   return unpack("f", pack("L", unpack("V", $f)));
}

sub
ddd'read_string
{
   local($string, $c);

   $string = "";
   while( read(DDD, $c, 1) )
   {
      last if $c eq "\000";
      $string = "$string$c";
   }
   return $string;
}
